/*
// CheckFrameCountAndAnimationSpeed.c	Last changed: 2022-04-18 by Jay
*/

#include <ctype.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define MAX_FILE_NAME_LENGTH 1000

static char CountFlag;
static unsigned char FatzHeader[8];
static unsigned char ExtraFrame;
static unsigned char AnimationSpeed[4];
static unsigned char LengthOfPalFileName[4];
static unsigned char PalFileName[256];
static unsigned char NumberOfFrames[4];
static int NumberOfWarnings;
static int NumberOfZTFiles;
static char QuietFlag;
static char VerboseFlag;

static void
checkDir(const char *dir)
{ int c;
  DIR *dirPtr;
  struct dirent *entryPtr;
  FILE *filePtr;
  long firstFrameCount;
  long firstSpeed;
  char foundBackslashFlag;
  long frameCount;
  int i;
  struct stat infoStruct;
  long speed;
  char ZTFileName[MAX_FILE_NAME_LENGTH];

  dirPtr = opendir(dir);

  if (dirPtr == NULL)
  { printf("***** Error: Could not open %s as directory.\n", dir);
    exit(1);
  }

  firstFrameCount = -1L;
  firstSpeed = -1L;

  while ((entryPtr = readdir(dirPtr)) != NULL)
  { if (strlen(dir) + strlen(entryPtr->d_name) + 2 > MAX_FILE_NAME_LENGTH)
    { printf("***** Error: Path %s/%s is too long.\n", dir, entryPtr->d_name);
      exit(2);
    }

    snprintf(ZTFileName, MAX_FILE_NAME_LENGTH,
             "%s/%s", dir, entryPtr->d_name);

    if (stat(ZTFileName, &infoStruct) != 0)
    { printf("***** Error: Cannot get info for %s.\n", ZTFileName);
      exit(3);
    }

    if ((infoStruct.st_mode & S_IFMT) == S_IFDIR)
    { if (strcmp(entryPtr->d_name, ".") != 0
          && strcmp(entryPtr->d_name, "..") != 0)
      { checkDir(ZTFileName);
      }
    }
    else if ((infoStruct.st_mode & S_IFMT) == S_IFREG)
    { if (strcasecmp(entryPtr->d_name, "e") == 0
          || strcasecmp(entryPtr->d_name, "n") == 0
          || strcasecmp(entryPtr->d_name, "ne") == 0
          || strcasecmp(entryPtr->d_name, "nw") == 0
          || strcasecmp(entryPtr->d_name, "s") == 0
          || strcasecmp(entryPtr->d_name, "se") == 0
          || strcasecmp(entryPtr->d_name, "sw") == 0
          || strcasecmp(entryPtr->d_name, "w") == 0)
      { filePtr = fopen(ZTFileName, "rb");

        if (filePtr == NULL)
        { printf("***** Error: File %s could not be opened.\n", ZTFileName);
          exit(4);
        }

        for (i = 0; i < 4; ++i)
        { if ((c = fgetc(filePtr)) == EOF)
          { printf("***** Error: File %s is not a ZT image file.\n",
                   ZTFileName);
            fclose(filePtr);
            exit(5);
          }

          FatzHeader[i] = c;
        }

        if (FatzHeader[0] == 'F' && FatzHeader[1] == 'A'
            && FatzHeader[2] == 'T' && FatzHeader[3] == 'Z')
        { for (i = 4; i < 8; ++i)
          { if ((c = fgetc(filePtr)) == EOF)
            { printf("***** Error: File %s is not a ZT image file.\n",
                     ZTFileName);
              fclose(filePtr);
              exit(6);
            }

            FatzHeader[i] = c;
          }

          if ((c = fgetc(filePtr)) == EOF)
          { printf("***** Error: File %s is not a ZT image file.\n",
                   ZTFileName);
            fclose(filePtr);
            exit(7);
          }

          ExtraFrame = c;

          for (i = 0; i < 4; ++i)
          { if ((c = fgetc(filePtr)) == EOF)
            { printf("***** Error: File %s is not a ZT image file.\n",
                     ZTFileName);
              fclose(filePtr);
              exit(8);
            }

            AnimationSpeed[i] = c;
          }
        }
        else
        { for (i = 0; i < 4; ++i)
          { AnimationSpeed[i] = FatzHeader[i];
            FatzHeader[i + 4] = 0;
          }

          FatzHeader[0] = 0;
          ExtraFrame = 0;
        }

        /*
        // The length for the ".pal" file is 4 bytes. But to speed things up
        // when this is not actually a ZT image, only the first byte is used.
        */

        for (i = 0; i < 4; ++i)
        { if ((c = fgetc(filePtr)) == EOF)
          { printf("***** Error: File %s is not a ZT image file.\n",
                   ZTFileName);
            fclose(filePtr);
            exit(9);
          }

          LengthOfPalFileName[i] = c;
        }

        if (LengthOfPalFileName[1] != 0 || LengthOfPalFileName[2] != 0
            || LengthOfPalFileName[3] != 0)
        { printf(
          "***** Error: File %s has pal file name length greater than 255.\n",
                 ZTFileName);
          fclose(filePtr);
          exit(10);
        }

        if (LengthOfPalFileName[0] < 6)
        { printf("***** Error: File %s is not a ZT image file.\n",
                 ZTFileName);
          fclose(filePtr);
          exit(11);
        }

        foundBackslashFlag = 'N';

        for (i = 0; i < LengthOfPalFileName[0]; ++i)
        { if ((c = fgetc(filePtr)) == EOF)
          { printf("***** Error: File %s is not a ZT image file.\n",
                   ZTFileName);
            fclose(filePtr);
            exit(12);
          }

          PalFileName[i] = c;

          if ( c == '\\')
          { foundBackslashFlag = 'Y';
          }
        }

        i = LengthOfPalFileName[0] - 5;

        if (PalFileName[i] != '.'
            || (PalFileName[i + 1] != 'p' && PalFileName[i + 1] != 'P')
            || (PalFileName[i + 2] != 'a' && PalFileName[i + 2] != 'A')
            || (PalFileName[i + 3] != 'l' && PalFileName[i + 3] != 'L')
            || PalFileName[i + 4] != '\0')
        { printf("***** Error: File %s is not a ZT image file.\n",
                 ZTFileName);
          fclose(filePtr);
          exit(13);
        }

        /*
        // The number of frames is 4 bytes.
        */

        for (i = 0; i < 4; ++i)
        { if ((c = fgetc(filePtr)) == EOF)
          { printf("***** Error: File %s is not a ZT image file.\n",
                   ZTFileName);
            fclose(filePtr);
            exit(14);
          }

          NumberOfFrames[i] = c;
        }

        fclose(filePtr);
        frameCount = 0L;
        speed = 0L;

        for (i = 3; i >=0; --i)
        { frameCount = (frameCount << 8) + (long)(NumberOfFrames[i]);
          speed = (speed << 8) + (long)(AnimationSpeed[i]);
        }

        if (VerboseFlag == 'Y')
        { printf("%s: Frame Count= %ld; Animation Speed= %ld\n",
                 ZTFileName, frameCount, speed);
        }

        ++NumberOfZTFiles;

        if (foundBackslashFlag == 'Y')
        { if (CountFlag != 'Y')
          { printf("*** Warning: pal path has backslashes in %s\n",
                   ZTFileName);
          }

          ++NumberOfWarnings;
        }

        if (firstFrameCount == -1L)
        { firstFrameCount = frameCount;
          firstSpeed = speed;
        }
        else
        { if (frameCount != firstFrameCount)
          { if (CountFlag != 'Y')
            { printf(
"*** Warning: Frame count %ld for %s does not match first frame count of %ld\n",
                     frameCount, ZTFileName, firstFrameCount);
            }

            ++NumberOfWarnings;
          }

          if (speed != firstSpeed
              && (frameCount != 1L || firstFrameCount != 1L))
          { if (CountFlag != 'Y')
            { printf(
"*** Warning: Speed %ld for %s does not match first animation speed of %ld\n",
                     speed, ZTFileName, firstSpeed);
            }

            ++NumberOfWarnings;
          }
        }
      }
    }
  }

  closedir(dirPtr);
  return;
}

int
main(int argc, char *argv[])
{ int i;

  CountFlag = 'N';
  NumberOfWarnings = 0;
  NumberOfZTFiles = 0;
  QuietFlag = 'N';
  VerboseFlag = 'N';

  for (i = 1; i < argc - 1; ++i)
  { if (strcmp(argv[i], "-c") == 0)
    { CountFlag = 'Y';
    }
    else if (strcmp(argv[i], "-q") == 0)
    { QuietFlag = 'Y';
    }
    else if (strcmp(argv[i], "-v") == 0)
    { VerboseFlag = 'Y';
    }
  }

  if (argc < 2 || (argc == 2 && *argv[1] == '-'))
  { printf("\n\
Usage:\n\
 %s [-c] [-q] [-v] {directory}\n\
\n\
This program checks for frame count and animation speed mismatches of all\n\
ZT image (e, n, ne, nw, s, se, sw, w) files in a specified directory\n\
and all of its subdirectories. This program also checks if backslashes\n\
were used for the pal file paths inside the ZT images.\n\
\n\
Adding \"-c\" will skip showing individual Warnings, but will still show\n\
how many warnings there were when there was at least 1.\n\
\n\
Adding \"-q\" will skip showing how many ZT image files were checked\n\
and will skip showing the \"There were no warnings\" message.\n\
\n\
Adding \"-v\" also shows the frame counts and animation speeds\n\
for all ZT image files.\n\n",
           argv[0]);
    exit(0);
  }

  checkDir(argv[argc - 1]);

  if (QuietFlag != 'Y')
  { if (NumberOfZTFiles == 0)
    { printf("There were no ZT image files.\n");
    }
    else if (NumberOfZTFiles == 1)
    { printf("Checked 1 ZT image file.\n");
    }
    else
    { printf("Checked %d ZT image files.\n", NumberOfZTFiles);
    }
  }

  if (NumberOfWarnings == 0)
  { if (QuietFlag != 'Y')
    { printf("There were no warnings for ZT image files.\n");
    }

    exit(0);
  }
  else if (NumberOfWarnings == 1)
  { printf("NOTE: There was 1 Warning for ZT image files.\n");
    exit(15);
  }

  else
  { printf("NOTE: There were %d Warnings for ZT image files.\n",
           NumberOfWarnings);
    exit(16);
  }

  return 0;
}

/* End of CheckFrameCountAndAnimationSpeed.c */
